import cgi
import hashlib
import json
import os

from kinto.authorization import RouteFactory
from kinto.core import utils as core_utils
from kinto.core.errors import raise_invalid
from kinto.core.storage import Filter
from kinto.views.records import Record
from pyramid import httpexceptions
from pyramid_storage.exceptions import FileNotAllowed


FILE_LINKS = "__attachments__"

RECORD_PATH = "/buckets/{bucket_id}/collections/{collection_id}/records/{id}"

DEFAULT_MIMETYPES = {
    ".pem": "application/x-pem-file",
    ".geojson": "application/geojson",
}


class AttachmentRouteFactory(RouteFactory):
    def __init__(self, request):
        """
        This class is the `context` object being passed to the
        :class:`kinto.core.authorization.AuthorizationPolicy`.

        Attachment is not a Kinto resource.

        The required permission is:
        * ``write`` if the related record exists;
        * ``record:create`` on the related collection otherwise.
        """
        super(AttachmentRouteFactory, self).__init__(request)
        self.resource_name = "record"
        try:
            request.current_resource_name = "record"
            request.validated.setdefault("header", {})
            request.validated.setdefault("querystring", {})
            resource = Record(request, context=self)
            resource.object_id = request.matchdict["id"]
            existing = resource.get()
        except httpexceptions.HTTPNotFound:
            existing = None

        if existing:
            # Request write permission on the existing record.
            self.permission_object_id = record_uri(request)
            self.required_permission = "write"
        else:
            # Request create record permission on the parent collection.
            self.permission_object_id = collection_uri(request)
            self.required_permission = "create"
        # Set the current object in context, since it is used in the
        # authorization policy to distinguish operations on plural endpoints
        # from individual objects. See Kinto/kinto#918
        self.current_object = existing


def sha256(content):
    m = hashlib.sha256()
    m.update(content)
    return m.hexdigest()


def _object_uri(request, resource_name, matchdict, prefix):
    uri = core_utils.instance_uri(request, resource_name=resource_name, **matchdict)
    if prefix:
        uri = f"/{request.registry.route_prefix}{uri}"
    return uri


def bucket_uri(request, prefix=False):
    matchdict = dict(request.matchdict)
    matchdict["id"] = matchdict["bucket_id"]
    return _object_uri(request, "bucket", matchdict, prefix)


def collection_uri(request, prefix=False):
    matchdict = dict(request.matchdict)
    matchdict["id"] = matchdict["collection_id"]
    return _object_uri(request, "collection", matchdict, prefix)


def record_uri(request, prefix=False):
    return _object_uri(request, "record", request.matchdict, prefix)


def patch_record(record, request):
    # XXX: add util clone_request()
    backup_pattern = request.matched_route.pattern
    backup_body = request.body
    backup_validated = request.validated

    # Instantiate record resource with current request.
    context = RouteFactory(request)
    context.resource_name = "record"
    context.get_permission_object_id = lambda r, i: record_uri(r)
    record_pattern = request.matched_route.pattern.replace("/attachment", "")
    request.matched_route.pattern = record_pattern

    # Simulate update of fields.
    request.validated = dict(body=record, **backup_validated)

    request.body = json.dumps(record).encode("utf-8")
    resource = Record(request, context=context)
    resource.object_id = request.matchdict["id"]
    setattr(request, "_attachment_auto_save", True)  # Flag in update listener.

    try:
        saved = resource.patch()
    except httpexceptions.HTTPNotFound:
        saved = resource.put()

    request.matched_route.pattern = backup_pattern
    request.body = backup_body
    request.validated = backup_validated
    return saved


def delete_attachment(request, link_field=None, uri=None, keep_old_files=False):
    """Delete existing file and link."""
    if link_field is None:
        link_field = "record_uri"
    if uri is None:
        uri = record_uri(request)

    storage = request.registry.storage
    filters = [Filter(link_field, uri, core_utils.COMPARISON.EQ)]

    # Remove file.
    if not keep_old_files:
        file_links = storage.list_all("", FILE_LINKS, filters=filters)
        for link in file_links:
            request.attachment.delete(link["location"])

    # Remove link.
    storage.delete_all("", FILE_LINKS, filters=filters, with_deleted=False)


def save_file(request, content, folder=None, keep_link=True, replace=False):
    randomize = setting_value(request, "randomize", default=True)

    overriden_mimetypes = {**DEFAULT_MIMETYPES}
    conf_mimetypes = setting_value(request, "mimetypes", default="")
    if conf_mimetypes:
        overriden_mimetypes.update(dict([v.split(":") for v in conf_mimetypes.split(";")]))

    # Read file to compute hash.
    if not isinstance(content, cgi.FieldStorage):
        error_msg = "Filename is required."
        raise_invalid(request, location="body", description=error_msg)

    # Posted file attributes.
    content.file.seek(0)
    filecontent = content.file.read()
    filehash = sha256(filecontent)
    size = len(filecontent)
    filename = content.filename

    _, extension = os.path.splitext(filename)
    mimetype = overriden_mimetypes.get(extension, content.type)

    save_options = {
        "folder": folder,
        "randomize": randomize,
        "replace": replace,
        "headers": {"Content-Type": mimetype},
    }

    try:
        location = request.attachment.save(content, **save_options)
    except FileNotAllowed:
        error_msg = "File extension is not allowed."
        raise_invalid(request, location="body", description=error_msg)

    # File metadata.
    fullurl = request.attachment.url(location)
    attachment = {
        "filename": filename,
        "location": fullurl,
        "hash": filehash,
        "mimetype": mimetype,
        "size": size,
    }

    if keep_link:
        # Store link between record and attachment (for later deletion).
        request.registry.storage.create(
            "",
            FILE_LINKS,
            {
                "location": location,  # store relative location.
                "bucket_uri": bucket_uri(request),
                "collection_uri": collection_uri(request),
                "record_uri": record_uri(request),
            },
        )

    return attachment


def setting_value(request, name, default):
    value = request.registry.settings.get("attachment.{}".format(name), default)
    if "bucket_id" in request.matchdict:
        uri = "/buckets/{bucket_id}".format(**request.matchdict)
        if uri in request.registry.attachment_resources:
            value = request.registry.attachment_resources[uri].get(name, value)
        if "collection_id" in request.matchdict:
            uri = "/buckets/{bucket_id}/collections/{collection_id}".format(**request.matchdict)
            if uri in request.registry.attachment_resources:
                value = request.registry.attachment_resources[uri].get(name, value)
    return value



from collections import defaultdict

import pkg_resources
from pyramid.exceptions import ConfigurationError
from pyramid.settings import asbool


#: Module version, as defined in PEP-0396.
__version__ = pkg_resources.get_distribution(__package__).version


def includeme(config):
    # Process settings to remove storage wording.
    settings = config.get_settings()

    storage_settings = {}
    config.registry.attachment_resources = defaultdict(dict)

    for setting_name, setting_value in settings.items():
        if setting_name.startswith("attachment."):
            if setting_name.startswith("attachment.resources."):
                # Resource specific config
                parts = setting_name.replace("attachment.resources.", "").split(".")
                # attachment.resources.{bid}.randomize
                # attachment.resources.{bid}.{cid}.randomize
                if len(parts) == 3:
                    bucket_id, collection_id, name = parts
                    resource_id = "/buckets/{}/collections/{}".format(bucket_id, collection_id)
                elif len(parts) == 2:
                    bucket_id, name = parts
                    resource_id = "/buckets/{}".format(bucket_id)
                else:
                    message = "Configuration rule malformed: `{}`".format(setting_name)
                    raise ConfigurationError(message)

                if name in ("randomize", "keep_old_files"):
                    config.registry.attachment_resources[resource_id][name] = asbool(setting_value)
                else:
                    message = "`{}` is not a supported setting name. Read `{}`".format(
                        name, setting_name
                    )
                    raise ConfigurationError(message)
            else:
                setting_name = setting_name.replace("attachment.", "storage.")
                storage_settings[setting_name] = setting_value

    # Force some pyramid_storage settings.
    storage_settings["storage.name"] = "attachment"
    storage_settings.setdefault("storage.extensions", "default")
    config.add_settings(storage_settings)

    # It may be useful to define an additional base_url setting.
    # (see workaround about relative base_url in README)
    extra_base_url = settings.get("attachment.extra.base_url", settings.get("attachment.base_url"))

    # Force trailing slash since attachment locations have no leading slash.
    if extra_base_url and not extra_base_url.endswith("/"):
        extra_base_url += "/"

    # Expose capability.
    config.add_api_capability(
        "attachments",
        version=__version__,
        description="Add file attachments to records",
        url="https://github.com/Kinto/kinto-attachment/",
        base_url=extra_base_url,
    )

    # Register heartbeat to check attachments storage.
    from kinto_attachment.views import attachments_ping

    config.registry.heartbeats["attachments"] = attachments_ping

    # Enable attachment backend.
    if "storage.base_path" in storage_settings:
        config.include("pyramid_storage.local")
    elif "storage.gcloud.credentials" in storage_settings:
        config.include("pyramid_storage.gcloud")
    else:
        config.include("pyramid_storage.s3")

    config.scan("kinto_attachment.views")
    config.scan("kinto_attachment.listeners")



from kinto.core.errors import http_error
from kinto.core.events import ACTIONS, ResourceChanged
from pyramid.events import subscriber
from pyramid.exceptions import HTTPBadRequest
from pyramid.settings import asbool

from . import utils


# XXX: Use AfterResourceChanged when implemented.
@subscriber(
    ResourceChanged,
    for_resources=("record", "collection", "bucket"),
    for_actions=(ACTIONS.DELETE,),
)
def on_delete_record(event):
    """When a resource record is deleted, delete all related attachments.
    When a bucket or collection is deleted, it removes the attachments of
    every underlying records.
    """
    keep_old_files = asbool(utils.setting_value(event.request, "keep_old_files", default=False))

    # Retrieve attachments for these records using links.
    resource_name = event.payload["resource_name"]
    filter_field = "%s_uri" % resource_name
    uri = event.payload["uri"]
    utils.delete_attachment(
        event.request, link_field=filter_field, uri=uri, keep_old_files=keep_old_files
    )


@subscriber(ResourceChanged, for_resources=("record",), for_actions=(ACTIONS.UPDATE,))
def on_update_record(event):
    if getattr(event.request, "_attachment_auto_save", False):
        # Record attributes are being updated by the plugin itself.
        return

    # A user is changing the record, make sure attachment metadata is not
    # altered manually.
    for change in event.impacted_objects:
        attachment_before = change["old"].get("attachment")
        attachment_after = change["new"].get("attachment")
        if attachment_before and attachment_after:
            if attachment_before != attachment_after:
                error_msg = "Attachment metadata cannot be modified."
                raise http_error(HTTPBadRequest(), message=error_msg)



import os
import unittest
import uuid
from unittest import mock
from urllib.parse import urlparse

from kinto.core.errors import ERRORS

from kinto_attachment.utils import sha256

from . import BaseWebTestLocal, BaseWebTestS3, get_user_headers


class UploadTest(object):
    def test_returns_200_to_record_once_uploaded(self):
        self.upload(status=201)

    def test_record_is_created_with_metadata(self):
        self.upload()
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertIn(self.file_field, resp.json["data"])

    def test_record_is_created_with_valid_id(self):
        self.record_uri = self.get_record_uri("fennec", "fonts", "logo")
        self.endpoint_uri = self.record_uri + "/attachment"
        self.app.put_json(self.record_uri, {}, headers=self.headers)
        self.upload(status=200)

    def test_returns_200_if_record_already_exists(self):
        self.app.put_json(self.record_uri, {}, headers=self.headers)
        self.upload(status=200)

    def test_adds_cors_and_location_to_response(self):
        response = self.upload()
        self.assertEqual(response.headers["Location"], "http://localhost/v1" + self.record_uri)
        self.assertIn("Access-Control-Allow-Origin", response.headers)

    def test_has_no_subfolder_if_setting_is_undefined(self):
        self.app.app.registry.settings.pop("attachment.folder")
        response = self.upload()
        record = self.get_record(response)
        url = urlparse(record["location"])
        self.assertNotIn("/", url.path[1:])

    def exists(self, fullurl):
        location = fullurl.replace(self.base_url, "")
        return self.backend.exists(location)

    def test_previous_attachment_is_removed_on_replacement(self):
        first = self.get_record(self.upload())
        self.assertTrue(self.exists(first["location"]))
        second = self.get_record(self.upload())
        self.assertFalse(self.exists(first["location"]))
        self.assertTrue(self.exists(second["location"]))


class LocalUploadTest(UploadTest, BaseWebTestLocal, unittest.TestCase):
    def test_file_is_created_on_local_filesystem(self):
        attachment = self.upload().json
        fullurl = attachment["location"]
        relativeurl = fullurl.replace(self.base_url, "")
        self.assertTrue(os.path.exists(os.path.join("/tmp", relativeurl)))

    def test_file_is_not_gzipped_on_local_filesystem(self):
        resp = self.upload(files=[(self.file_field, b"my-report.pdf", b"--binary--")])
        attachment = resp.json
        self.assertTrue(attachment["location"].endswith(".pdf"))
        self.assertEqual(attachment["mimetype"], "application/pdf")
        relativeurl = attachment["location"].replace(self.base_url, "")
        self.assertEqual(attachment["hash"], sha256(b"--binary--"))
        self.assertEqual(attachment["size"], len(b"--binary--"))
        file_path = os.path.join("/tmp", relativeurl)
        self.assertTrue(os.path.exists(file_path))
        with open(file_path, "rb") as f:
            self.assertEqual(f.read(), b"--binary--")


class S3UploadTest(UploadTest, BaseWebTestS3, unittest.TestCase):
    pass


class DeleteTest(object):
    def setUp(self):
        super(DeleteTest, self).setUp()
        self.attachment = self.upload().json

    def exists(self, fullurl):
        location = fullurl.replace(self.base_url, "")
        return self.backend.exists(location)

    def test_attachment_is_removed_on_delete(self):
        fullurl = self.attachment["location"]
        self.assertTrue(self.exists(fullurl))
        self.app.delete(self.endpoint_uri, headers=self.headers, status=204)
        self.assertFalse(self.exists(fullurl))

    def test_metadata_are_removed_on_delete(self):
        self.app.delete(self.endpoint_uri, headers=self.headers, status=204)
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertIsNone(resp.json["data"].get("attachment"))

    def test_link_is_removed_on_delete(self):
        storage = self.app.app.registry.storage
        links = storage.list_all("", "__attachments__")
        self.assertEqual(len(links), self.nb_uploaded_files)
        self.app.delete(self.endpoint_uri, headers=self.headers, status=204)
        links = storage.list_all("", "__attachments__")
        self.assertEqual(len(links), 0)

    def test_attachment_is_removed_when_record_is_deleted(self):
        fullurl = self.attachment["location"]
        self.assertTrue(self.exists(fullurl))
        self.app.delete(self.record_uri, headers=self.headers)
        self.assertFalse(self.exists(fullurl))

    def test_attachments_are_removed_when_bucket_is_deleted(self):
        fullurl = self.attachment["location"]
        self.assertTrue(self.exists(fullurl))
        self.app.delete("/buckets/fennec", headers=self.headers)
        self.assertFalse(self.exists(fullurl))

    def test_attachments_are_removed_when_collection_is_deleted(self):
        fullurl = self.attachment["location"]
        self.assertTrue(self.exists(fullurl))
        self.app.delete("/buckets/fennec/collections/fonts", headers=self.headers)
        self.assertFalse(self.exists(fullurl))

    def test_attachments_links_are_removed_forever(self):
        storage = self.app.app.registry.storage
        links = storage.list_all("", "__attachments__")
        self.assertEqual(len(links), self.nb_uploaded_files)
        self.app.delete(self.record_uri, headers=self.headers)
        links = storage.list_all("", "__attachments__")
        self.assertEqual(len(links), 0)

    def test_no_error_when_other_resource_is_deleted(self):
        group_url = "/buckets/default/groups/admins"
        self.app.put_json(group_url, {"data": {"members": ["them"]}}, headers=self.headers)
        self.app.delete(group_url, headers=self.headers)


class LocalDeleteTest(DeleteTest, BaseWebTestLocal, unittest.TestCase):
    pass


class S3DeleteTest(DeleteTest, BaseWebTestS3, unittest.TestCase):
    pass


class AttachmentViewTest(object):
    def test_only_post_and_options_is_accepted(self):
        self.app.get(self.endpoint_uri, headers=self.headers, status=405)
        self.app.put(self.endpoint_uri, headers=self.headers, status=405)
        self.app.patch(self.endpoint_uri, headers=self.headers, status=405)
        headers = self.headers.copy()
        headers["Access-Control-Request-Method"] = "POST"
        self.app.options(self.endpoint_uri, headers=headers, status=200)

    def test_record_is_updated_with_metadata(self):
        existing = {"data": {"author": "frutiger"}}
        self.app.put_json(self.record_uri, existing, headers=self.headers)
        self.upload()
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertIn(self.file_field, resp.json["data"])
        self.assertIn("author", resp.json["data"])

    def test_record_metadata_has_hash_hexdigest(self):
        r = self.upload()
        h = "db511d372e98725a61278e90259c7d4c5484fc7a781d7dcc0c93d53b8929e2ba"
        self.assertEqual(self.get_record(r)["hash"], h)

    def test_record_metadata_has_randomized_location(self):
        resp = self.upload(files=[(self.file_field, b"my-report.pdf", b"--binary--")])
        record = self.get_record(resp)
        self.assertNotIn("report", record["location"])

    def test_record_location_contains_subfolder(self):
        self.upload()
        resp = self.app.get(self.record_uri, headers=self.headers)
        location = resp.json["data"][self.file_field]["location"]
        self.assertIn("fennec/fonts/", location)

    def test_record_metadata_provides_original_filename(self):
        resp = self.upload(files=[(self.file_field, b"my-report.pdf", b"--binary--")])
        record = self.get_record(resp)
        self.assertEqual("my-report.pdf", record["filename"])

    def test_record_is_created_with_fields(self):
        self.upload(params=[("data", '{"family": "sans"}')])
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertEqual(resp.json["data"]["family"], "sans")

    def test_record_is_updated_with_fields(self):
        existing = {"data": {"author": "frutiger"}}
        self.app.put_json(self.record_uri, existing, headers=self.headers)
        self.upload(params=[("data", '{"family": "sans"}')])
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertEqual(resp.json["data"]["family"], "sans")
        self.assertEqual(resp.json["data"]["author"], "frutiger")

    def test_record_attachment_metadata_cannot_be_removed_manually(self):
        self.upload(params=[("data", '{"family": "sans"}')])
        body = {"data": {"attachment": {"manual": "true"}}}
        resp = self.app.patch_json(self.record_uri, body, headers=self.headers, status=400)
        self.assertIn("Attachment metadata cannot be modified", resp.json["message"])

    def test_record_is_created_with_appropriate_permissions(self):
        self.upload()
        current_principal = (
            "basicauth:c6c27f0c7297ba7d4abd2a70c8a2cb88a06a3" "bb793817ef2c85fe8a709b08022"
        )
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertEqual(resp.json["permissions"], {"write": [current_principal]})

    def test_record_permissions_can_also_be_specified(self):
        self.upload(params=[("permissions", '{"read": ["system.Everyone"]}')])
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertIn("system.Everyone", resp.json["permissions"]["read"])

    # Content Validation.

    def test_records_fields_must_be_valid_json(self):
        resp = self.upload(params=[("data", "{>author: 12}")], status=400)
        self.assertIn("body: data is not valid JSON", resp.json["message"])

    def test_permissions_must_be_valid_json(self):
        resp = self.upload(params=[("permissions", '{"read": >}')], status=400)
        self.assertIn("body: permissions is not valid JSON", resp.json["message"])

    def test_permissions_must_be_str(self):
        files = [
            ("attachment", "image.jpg", b"--fake--"),
            ("data", "data", b'{"family": "sans"}'),
            ("permissions", "permissions", b'{"read": ["system.Everyone"]}'),
        ]
        content_type, body = self.app.encode_multipart([], files)
        headers = {**self.headers, "Content-Type": content_type}

        resp = self.app.post(self.endpoint_uri, body, headers=headers, status=400)

        self.assertIn(
            "body: 'data' field should be passed as form data, not files", resp.json["message"]
        )

    def test_unknown_fields_are_not_accepted(self):
        resp = self.upload(params=[("my_field", "a_value")], status=400)
        self.assertIn("body: 'my_field' not in ('data', 'permissions')", resp.json["message"])

    def test_record_fields_are_validated_against_schema(self):
        resp = self.upload(params=[("data", '{"author": 12}')], status=400)
        self.assertIn("author in body: 12 is not of type ", resp.json["message"])

    def test_attachment_must_have_a_filename(self):
        resp = self.upload(files=[(self.file_field, b"", b"--fake--")], status=400)
        self.assertEqual(resp.json["message"], "body: Filename is required.")

    def test_upload_refused_if_extension_not_allowed(self):
        resp = self.upload(files=[(self.file_field, b"virus.exe", b"--fake--")], status=400)
        self.assertEqual(resp.json["message"], "body: File extension is not allowed.")

    def test_upload_refused_if_field_is_not_attachment(self):
        resp = self.upload(files=[("fichierjoint", b"image.jpg", b"--fake--")], status=400)
        self.assertEqual(resp.json["message"], "Attachment missing.")
        self.assertEqual(resp.json["errno"], ERRORS.INVALID_POSTED_DATA.value)

    def test_upload_refused_if_header_is_not_multipart(self):
        self.headers["Content-Type"] = "application/json"
        resp = self.app.post(self.endpoint_uri, {}, headers=self.headers, status=400)
        self.assertEqual(resp.json["message"], "Content-Type should be multipart/form-data")
        self.assertEqual(resp.json["errno"], ERRORS.INVALID_PARAMETERS.value)

    def test_upload_refused_if_header_is_invalid_multipart(self):
        self.headers["Content-Type"] = "multipart/form-data"
        resp = self.app.post(self.endpoint_uri, {}, headers=self.headers, status=400)
        self.assertEqual(
            resp.json["message"].replace(": b'", ": '"), "Invalid boundary in multipart form: ''"
        )
        self.assertEqual(resp.json["errno"], ERRORS.INVALID_PARAMETERS.value)

    # Permissions.

    def test_upload_refused_if_not_authenticated(self):
        self.headers.pop("Authorization")
        self.upload(status=401)

    def test_upload_replace_refused_if_not_authenticated(self):
        self.upload(status=201)

        self.headers.pop("Authorization")
        self.upload(status=401)

    def test_upload_refused_if_not_allowed(self):
        self.headers.update(get_user_headers("jean-louis"))
        self.upload(status=403)

    def test_upload_replace_refused_if_only_create_allowed(self):
        # Allow any authenticated to write in this collection.
        perm = {"permissions": {"record:create": ["system.Authenticated"]}}
        self.app.patch_json("/buckets/fennec/collections/fonts", perm, headers=self.headers)
        self.upload(status=201)

        self.headers.update(get_user_headers("jean-louis"))
        self.upload(status=403)

    def test_upload_replace_refused_if_only_bucket_read_is_allowed(self):
        # Create a record with attachment.
        self.upload(status=201)

        # Now allow anyone to read this bucket.
        perm = {"permissions": {"read": ["system.Everyone"]}}
        self.app.patch_json("/buckets/fennec", perm, headers=self.headers)

        # And try to replace anonymously.
        self.headers.pop("Authorization")
        self.upload(status=401)

    def test_upload_replace_refused_if_only_read_is_allowed(self):
        # Create a record with attachment.
        self.upload(status=201)

        # Now allow anyone to read this collection.
        perm_change = [
            {"op": "add", "path": "/permissions", "value": {"read": ["system.Everyone"]}}
        ]
        self.app.patch_json(
            "/buckets/fennec/collections/fonts",
            perm_change,
            headers={**self.headers, "Content-Type": "application/json-patch+json"},
        )

        # And try to replace anonymously.
        self.headers.pop("Authorization")
        self.upload(status=401)

    def test_upload_create_accepted_if_create_allowed(self):
        # Allow any authenticated to write in this collection.
        perm = {"permissions": {"record:create": ["system.Authenticated"]}}
        self.app.patch_json("/buckets/fennec/collections/fonts", perm, headers=self.headers)

        self.headers.update(get_user_headers("jean-louis"))
        self.upload(status=201)

    def test_upload_create_accepted_if_write_allowed(self):
        # Allow any authenticated to write in this bucket.
        perm = {"permissions": {"write": ["system.Authenticated"]}}
        self.app.patch_json("/buckets/fennec", perm, headers=self.headers)

        self.headers.update(get_user_headers("jean-louis"))
        self.upload(status=201)

    def test_upload_replace_accepted_if_write_allowed(self):
        # Allow any authenticated to write in this bucket.
        perm = {"permissions": {"write": ["system.Authenticated"]}}
        self.app.patch_json("/buckets/fennec", perm, headers=self.headers)
        self.upload(status=201)

        self.headers.update(get_user_headers("jean-louis"))
        self.upload(status=200)


class PerResourceConfigAttachementViewTest(BaseWebTestS3, unittest.TestCase):
    config = "config/s3_per_resource.ini"

    def test_file_get_randomize_in_fennec_bucket(self):
        r = self.upload()

        self.assertEqual(r.json["filename"], "image.jpg")
        self.assertNotIn(r.json["filename"], r.json["location"])

    def test_file_do_not_get_randomize_in_fennec_experiments_collection(self):
        self.create_collection("fennec", "experiments")
        record_uri = self.get_record_uri("fennec", "experiments", str(uuid.uuid4()))
        self.endpoint_uri = record_uri + "/attachment"
        r = self.upload()

        self.assertEqual(r.json["filename"], "image.jpg")
        self.assertIn(r.json["filename"], r.json["location"])


class OverridenMimetypesTest(BaseWebTestS3, unittest.TestCase):
    config = "config/s3.ini"

    def test_file_mimetype_comes_from_config(self):
        resp = self.upload(files=[(self.file_field, b"kinto.txt", b"--binary--")])

        self.assertEqual(resp.json["mimetype"], "text/vnd.graphviz")


class SingleAttachmentViewTest(AttachmentViewTest, BaseWebTestLocal, unittest.TestCase):
    pass


class DefaultBucketTest(BaseWebTestLocal, unittest.TestCase):
    def setUp(self):
        super(DefaultBucketTest, self).setUp()
        self.record_uri = self.get_record_uri("default", "pix", uuid.uuid4())
        self.endpoint_uri = self.record_uri + "/attachment"

    def test_implicit_collection_creation_on_upload(self):
        resp = self.upload()
        record_uri = resp.headers["Location"]
        self.assertIn("/buckets/b4a52ebc-fe4a-1167-89f3-c792640c70b3", record_uri)


class KeepOldFilesTest(BaseWebTestLocal, unittest.TestCase):
    def make_app(self):
        import webtest
        from kinto import DEFAULT_SETTINGS
        from kinto import main as testapp
        from kinto.core import testing as core_support

        settings = core_support.DEFAULT_SETTINGS.copy()
        settings.update(**DEFAULT_SETTINGS)
        settings["multiauth.policies"] = "basicauth"
        settings["storage_backend"] = "kinto.core.storage.memory"
        settings["permission_backend"] = "kinto.core.permission.memory"
        settings["userid_hmac_secret"] = "this is not a secret"
        settings["includes"] = "kinto_attachment"

        settings["kinto.attachment.base_path"] = "/tmp"
        settings["kinto.attachment.base_url"] = ""
        settings["kinto.attachment.keep_old_files"] = "true"

        app = webtest.TestApp(testapp({}, **settings))
        app.RequestClass = core_support.get_request_class(prefix="v1")
        return app

    def test_files_are_kept_when_attachment_is_replaced(self):
        resp = self.upload(status=201)
        location1 = resp.json["location"]
        resp = self.upload(status=200)
        location2 = resp.json["location"]
        self.assertNotEqual(location1, location2)
        self.assertTrue(self.backend.exists(location2))
        self.assertTrue(self.backend.exists(location1))

    def test_files_are_kept_when_attachment_is_deleted(self):
        resp = self.upload(status=201)
        location = resp.json["location"]
        self.assertTrue(self.backend.exists(location))

        self.app.delete(self.record_uri + "/attachment", headers=self.headers)

        self.assertTrue(self.backend.exists(location))

    def test_files_are_kept_when_record_is_deleted(self):
        resp = self.upload(status=201)
        location = resp.json["location"]
        self.assertTrue(self.backend.exists(location))

        self.app.delete(self.record_uri, headers=self.headers)

        self.assertTrue(self.backend.exists(location))


class HeartbeartTest(BaseWebTestS3, unittest.TestCase):
    def test_attachments_is_added_to_heartbeat_view(self):
        resp = self.app.get("/__heartbeat__")
        self.assertIn("attachments", resp.json)

    def test_heartbeat_is_false_if_error_happens(self):
        with mock.patch("pyramid_storage.s3.S3FileStorage.delete") as mocked:
            mocked.side_effect = ValueError
            resp = self.app.get("/__heartbeat__", status=503)
        self.assertFalse(resp.json["attachments"])

    def test_heartbeat_is_true_if_server_is_readonly(self):
        patch = mock.patch("pyramid_storage.s3.S3FileStorage.delete")
        self.addCleanup(patch.stop)
        mocked = patch.start()
        mocked.side_effect = ValueError

        with mock.patch.dict(self.app.app.registry.settings, [("readonly", "true")]):
            resp = self.app.get("/__heartbeat__")
        self.assertTrue(resp.json["attachments"])
